//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//  This file is part of the buola project (https://code.google.com/p/buola/).  //
//                                                                              //
//  Copyright(c) 2007-2013 Xavi Gratal                                          //
//  gratal AT gmail DOT com                                                     //
//                                                                              //
//  Buola is free software: you can redistribute it and/or modify               //
//  it under the terms of the GNU General Public License as published by        //
//  the Free Software Foundation, either version 3 of the License, or           //
//  (at your option) any later version.                                         //
//                                                                              //
//  Buola is distributed in the hope that it will be useful,                    //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of              //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               //
//  GNU General Public License for more details.                                //
//                                                                              //
//  You should have received a copy of the GNU General Public License           //
//  along with buola.  If not, see <http://www.gnu.org/licenses/>.              //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

#ifndef _BUOLA_FUNCTORS_PREDICATES_CHAR_H_
#define _BUOLA_FUNCTORS_PREDICATES_CHAR_H_

#include <buola/functors/predicates/unary.h>

namespace buola { namespace fn {

///\addtogroup functors
///@{

///\name Character classification predicates
//@{

/////////////////////////////////////////////////////////////////////////////////////////////////
///
/// ctype-check predicate
///
/// Evaluated with a single argument, returns true if the parameter belongs to one of the ctypes
/// specified in the constructor, for the given locale
///
/// This class caches the ctype facets for the given locale, both for char and wchar_t, and can
/// thus be used with any of those types. The rationale for that is that obtaining the facet
/// from the locale is a much more expensive operation than actually using the facet to check
/// the type of a particular character.
///
/// The default copy constructor can be used since even that the locale object will be different,
/// the facets they point to will be identical.
///
/////////////////////////////////////////////////////////////////////////////////////////////////
struct AIsOfType : public GUnaryPredicate
{
    ///constructs the predicate given a ctype mask and a locale, which defaults to the current one
    explicit AIsOfType(std::ctype_base::mask pType,const std::locale &pLocale=std::locale())
        :   mLocale(pLocale)
        ,   mType(pType)
        ,   mCharFacet(nullptr)
        ,   mWCharFacet(nullptr)
    {}
    
    ///returns true if the character is of the given type
    bool operator()(char pCh) const
    {
        if(!mCharFacet) mCharFacet=&std::use_facet<std::ctype<char>>(mLocale);
        return mCharFacet->is(mType,pCh);
    }
    
    ///returns true if the character is of the given type
    bool operator()(wchar_t pCh) const
    {
        if(!mWCharFacet) mWCharFacet=&std::use_facet<std::ctype<wchar_t>>(mLocale);
        return mWCharFacet->is(mType,pCh);
    }
    
private:
    std::locale mLocale;
    std::ctype_base::mask mType;
    mutable const std::ctype<char> *mCharFacet;
    mutable const std::ctype<wchar_t> *mWCharFacet;
};

/////////////////////////////////////////////////////////////////////////////////////////////////
///
/// ctype-check predicate (typed version)
///
/// Evaluated with a single argument, returns true if the parameter belongs to one of the ctypes
/// specified in the constructor, for the given locale
///
/// This class differs from \ref AIsOfType in that it uses the facet for the type provided as a
/// template parameter
///
/////////////////////////////////////////////////////////////////////////////////////////////////
template<typename tChar>
struct AIsOfTypeTyped : public GUnaryPredicate
{
    ///constructs the predicate given a ctype mask and a locale, which defaults to the current one
    explicit AIsOfTypeTyped(std::ctype_base::mask pType,const std::locale &pLocale=std::locale())
        :   mLocale(pLocale)
        ,   mFacet(std::use_facet<std::ctype<tChar>>(pLocale))
        ,   mType(pType)
    {}
    
    ///returns true if the character is of the given type
    bool operator()(tChar pCh) const
    {
        return mFacet.is(mType,pCh);
    }
    
private:
    std::locale mLocale;
    const std::ctype<tChar> &mFacet;
    std::ctype_base::mask mType;
};

///returns a predicate that returns true if its argument is of the given character type
inline AIsOfType is_of_ctype(std::ctype_base::mask pType,const std::locale &pLocale=std::locale())
{
    return AIsOfType(pType,pLocale);
}

///returns a predicate that returns true if its argument is of the given character type (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_of_ctype(std::ctype_base::mask pType,const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(pType,pLocale);
}

///returns a predicate that returns true if its argument is alphanumeric
inline AIsOfType is_alnum(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::alnum,pLocale);
}

///returns a predicate that returns true if its argument is alphanumeric (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_alnum(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::alnum,pLocale);
}

///returns a predicate that returns true if its argument is a letter
inline AIsOfType is_alpha(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::alpha,pLocale);
}

///returns a predicate that returns true if its argument is a letter (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_alpha(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::alpha,pLocale);
}

///returns a predicate that returns true if its argument is a control character
inline AIsOfType is_cntrl(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::cntrl,pLocale);
}

///returns a predicate that returns true if its argument is a control character (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_cntrl(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::cntrl,pLocale);
}

///returns a predicate that returns true if its argument is a digit
inline AIsOfType is_digit(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::digit,pLocale);
}

///returns a predicate that returns true if its argument is a digit (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_digit(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::digit,pLocale);
}

///returns a predicate that returns true if its argument is a graphic character
inline AIsOfType is_graph(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::graph,pLocale);
}

///returns a predicate that returns true if its argument is a graphic character (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_graph(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::graph,pLocale);
}

///returns a predicate that returns true if its argument is a lowercase letter
inline AIsOfType is_lower(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::lower,pLocale);
}

///returns a predicate that returns true if its argument is a lowercase letter (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_lower(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::lower,pLocale);
}

///returns a predicate that returns true if its argument is a printable character
inline AIsOfType is_print(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::print,pLocale);
}

///returns a predicate that returns true if its argument is a printable character (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_print(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::print,pLocale);
}

///returns a predicate that returns true if its argument is a punctuation symbol
inline AIsOfType is_punct(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::punct,pLocale);
}

///returns a predicate that returns true if its argument is a punctuation symbol (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_punct(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::punct,pLocale);
}

///returns a predicate that returns true if its argument is a space
inline AIsOfType is_space(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::space,pLocale);
}

///returns a predicate that returns true if its argument is a space (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_space(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::space,pLocale);
}

///returns a predicate that returns true if its argument is an uppercase letter
inline AIsOfType is_upper(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::upper,pLocale);
}

///returns a predicate that returns true if its argument is an uppercase letter (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_upper(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::upper,pLocale);
}

///returns a predicate that returns true if its argument is an hexadecimal digit (typed version)
inline AIsOfType is_xdigit(const std::locale &pLocale=std::locale())
{
    return AIsOfType(std::ctype_base::xdigit,pLocale);
}

///returns a predicate that returns true if its argument is an hexadecimal digit (typed version)
template<typename tType>
inline AIsOfTypeTyped<tType> is_xdigit(const std::locale &pLocale=std::locale())
{
    return AIsOfTypeTyped<tType>(std::ctype_base::xdigit,pLocale);
}

//@}

///@}

/*namespace fn*/ } /*namespace buola*/ }

#endif
