//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//  This file is part of the buola project (https://code.google.com/p/buola/).  //
//                                                                              //
//  Copyright(c) 2007-2012 Xavi Gratal                                          //
//  gratal AT gmail DOT com                                                     //
//                                                                              //
//  Buola is free software: you can redistribute it and/or modify               //
//  it under the terms of the GNU General Public License as published by        //
//  the Free Software Foundation, either version 3 of the License, or           //
//  (at your option) any later version.                                         //
//                                                                              //
//  Buola is distributed in the hope that it will be useful,                    //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of              //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               //
//  GNU General Public License for more details.                                //
//                                                                              //
//  You should have received a copy of the GNU General Public License           //
//  along with buola.  If not, see <http://www.gnu.org/licenses/>.              //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

#include <buola/buola.h>
#include <buola/datetime/cdatetime.h>

#include <cwchar>
#include <cstdlib>
#include <ctime>
#include <sstream>

namespace buola {

///////////////////////////////////// integral types ///////////////////////////////////

////////////////////////////////////////////////////////////////
// int8_t (it's also signed char, and is converted as number)
//

template<>
int8_t semantic_cast<int8_t,std::string>(const std::string &pStr)
{
    return (int8_t)std::stol(pStr);
}

template<>
std::string semantic_cast<std::string,int8_t>(const int8_t &pByte)
{
    return std::to_string((int32_t)pByte);
}

template<>
int8_t semantic_cast<int8_t,std::wstring>(const std::wstring &pStr)
{
    return (int8_t)std::stol(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,int8_t>(const int8_t &pByte)
{
    return std::to_wstring((int32_t)pByte);
}

///////////////////////////////////////////////////////////////////
// uint8_t (it's also unsigned char, and is converted as number)
//

template<>
uint8_t semantic_cast<uint8_t,std::string>(const std::string &pStr)
{
    return (uint8_t)std::stoul(pStr);
}

template<>
std::string semantic_cast<std::string,uint8_t>(const uint8_t &pByte)
{
    return std::to_string((uint32_t)pByte);
}

template<>
uint8_t semantic_cast<uint8_t,std::wstring>(const std::wstring &pStr)
{
    return (uint8_t)std::stoul(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,uint8_t>(const uint8_t &pByte)
{
    return std::to_wstring((uint32_t)pByte);
}

/////////////
// int16_t
//

template<>
int16_t semantic_cast<int16_t,std::string>(const std::string &pStr)
{
    return (int16_t)std::stol(pStr);
}

template<>
std::string semantic_cast<std::string,int16_t>(const int16_t &p16)
{
    return std::to_string((int32_t)p16);
}

template<>
int16_t semantic_cast<int16_t,std::wstring>(const std::wstring &pStr)
{
    return (int16_t)std::stol(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,int16_t>(const int16_t &p16)
{
    return std::to_wstring((int32_t)p16);
}

//////////////
// uint16_t
//

template<>
uint16_t semantic_cast<uint16_t,std::string>(const std::string &pStr)
{
    return (uint16_t)std::stoul(pStr);
}

template<>
std::string semantic_cast<std::string,uint16_t>(const uint16_t &p16)
{
    return std::to_string((uint32_t)p16);
}

template<>
uint16_t semantic_cast<uint16_t,std::wstring>(const std::wstring &pStr)
{
    return (uint16_t)std::stoul(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,uint16_t>(const uint16_t &p16)
{
    return std::to_wstring((uint32_t)p16);
}

/////////////
// int32_t
//

template<>
int32_t semantic_cast<int32_t,std::string>(const std::string &pStr)
{
    return (int32_t)std::stol(pStr);
}

template<>
std::string semantic_cast<std::string,int32_t>(const int32_t &p32)
{
    return std::to_string(p32);
}

template<>
int32_t semantic_cast<int32_t,std::wstring>(const std::wstring &pStr)
{
    return (int32_t)std::stol(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,int32_t>(const int32_t &p32)
{
    return std::to_wstring(p32);
}

//////////////
// uint32_t
//

template<>
uint32_t semantic_cast<uint32_t,std::string>(const std::string &pStr)
{
    return (uint32_t)std::stoul(pStr);
}

template<>
std::string semantic_cast<std::string,uint32_t>(const uint32_t &p32)
{
    return std::to_string(p32);
}

template<>
uint32_t semantic_cast<uint32_t,std::wstring>(const std::wstring &pStr)
{
    return (uint32_t)std::stoul(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,uint32_t>(const uint32_t &p32)
{
    return std::to_wstring(p32);
}

/////////////
// int64_t
//

template<>
int64_t semantic_cast<int64_t,std::string>(const std::string &pStr)
{
    return (int64_t)std::stol(pStr);
}

template<>
std::string semantic_cast<std::string,int64_t>(const int64_t &p64)
{
    return std::to_string(p64);
}

template<>
int64_t semantic_cast<int64_t,std::wstring>(const std::wstring &pStr)
{
    return (int64_t)std::stol(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,int64_t>(const int64_t &p64)
{
    return std::to_wstring(p64);
}

//////////////
// uint64_t
//

template<>
uint64_t semantic_cast<uint64_t,std::string>(const std::string &pStr)
{
    return (uint64_t)std::stoul(pStr);
}

template<>
std::string semantic_cast<std::string,uint64_t>(const uint64_t &p64)
{
    return std::to_string(p64);
}

template<>
uint64_t semantic_cast<uint64_t,std::wstring>(const std::wstring &pStr)
{
    return (uint64_t)std::stoul(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,uint64_t>(const uint64_t &p64)
{
    return std::to_wstring(p64);
}

#if BUOLA_PLATFORM_MAC

//////////////
//size_t (it's long, while uint64_t is long long, so different types)
//

template<>
size_t semantic_cast<size_t,std::string>(const std::string &pStr)
{
    return (size_t)std::stoul(pStr);
}

template<>
std::string semantic_cast<std::string,size_t>(const size_t &p64)
{
    return std::to_string(p64);
}

template<>
size_t semantic_cast<size_t,std::wstring>(const std::wstring &pStr)
{
    return (size_t)std::stoul(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,size_t>(const size_t &p64)
{
    return std::to_wstring(p64);
}

#endif //BUOLA_PLATFORM_MAC

/////////////////////////////// floating point types /////////////////////////////

////////////
// double
//

template<>
double semantic_cast<double,std::string>(const std::string &pStr)
{
    return std::stod(pStr);
}

template<>
std::string semantic_cast<std::string,double>(const double &pDouble)
{
    return std::to_string(pDouble);
}

template<>
double semantic_cast<double,std::wstring>(const std::wstring &pStr)
{
    return std::stod(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,double>(const double &pDouble)
{
    return std::to_wstring(pDouble);
}

///////////
// float
//

template<>
float semantic_cast<float,std::string>(const std::string &pStr)
{
    return std::stof(pStr);
}

template<>
std::string semantic_cast<std::string,float>(const float &pFloat)
{
    return std::to_string(pFloat);
}

template<>
float semantic_cast<float,std::wstring>(const std::wstring &pStr)
{
    return std::stod(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,float>(const float &pFloat)
{
    return std::to_wstring(pFloat);
}

/////////////////
// long double
//

template<>
long double semantic_cast<long double,std::string>(const std::string &pStr)
{
    return std::stold(pStr);
}

template<>
std::string semantic_cast<std::string,long double>(const long double &pDouble)
{
    return std::to_string(pDouble);
}

template<>
long double semantic_cast<long double,std::wstring>(const std::wstring &pStr)
{
    return std::stold(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,long double>(const long double &pDouble)
{
    return std::to_wstring(pDouble);
}

////////////////////////////////// special types /////////////////////////////////

///////////////////////////////////////////////
// wchar_t (this is converted as a character)
//

template<>
wchar_t semantic_cast<wchar_t,std::string>(const std::string &pStr)
{
    return decode(pStr)[0];
}

template<>
std::string semantic_cast<std::string,wchar_t>(const wchar_t &pChar)
{
    std::wstring lRet;
    lRet+=pChar;
    return encode(lRet);
}

template<>
wchar_t semantic_cast<wchar_t,std::wstring>(const std::wstring &pStr)
{
    return pStr[0];
}

template<>
std::wstring semantic_cast<std::wstring,wchar_t>(const wchar_t &pChar)
{
    std::wstring lRet;
    lRet+=pChar;
    return lRet;
}

//////////
// bool
//

template<>
bool semantic_cast<bool,std::string>(const std::string &pStr)
{
    if(!pStr.length()) return false;
    
    if(pStr[0]=='f'||pStr[0]=='F'||pStr[0]=='n'||
        pStr[0]=='N'||pStr=="0")
    {
        return false;
    }
    else
    {
        return true;
    }
}

template<>
std::string semantic_cast<std::string,bool>(const bool &pBool)
{
    if(pBool)
        return "true";
    else
        return "false";
}

template<>
bool semantic_cast<bool,std::wstring>(const std::wstring &pStr)
{
    if(pStr[0]=='f'||pStr[0]=='F'||pStr[0]=='n'||
        pStr[0]=='N'||pStr==L"0")
    {
        return false;
    }
    else
    {
        return true;
    }
}

template<>
std::wstring semantic_cast<std::wstring,bool>(const bool &pBool)
{
    if(pBool)
        return L"true";
    else
        return L"false";
}

/////////////////////////////////////////////////////////////////////////////
// void* (printed as an hexadecimal number, no conversion back to pointer)
//

template<>
void* semantic_cast<void*,std::string>(const std::string& /*pStr*/)
{
    return nullptr;
}

template<>
std::string semantic_cast<std::string,void*>(void * const &pPointer)
{
    char lBuffer[100];
    snprintf(lBuffer,100,"%p",pPointer);
    
    return lBuffer;
}

template<>
void* semantic_cast<void*,std::wstring>(const std::wstring& /*pStr*/)
{
    return nullptr;
}

template<>
std::wstring semantic_cast<std::wstring,void*>(void * const &pPointer)
{
    wchar_t lBuffer[100];
    swprintf(lBuffer,100,L"%p",pPointer);
    
    return lBuffer;
}

///////////////////////////////// datetime classes ////////////////////////////////
    
///////////
// CTime
    
template<>
CCalendarTime semantic_cast<CCalendarTime,std::string>(const std::string &pStr)
{
    struct tm lTM;

    strptime(pStr.c_str(),"%T %F %z",&lTM);

    return CCalendarClock::from_time_t(mktime(&lTM));
}

template<>
std::string semantic_cast<std::string,CCalendarTime>(const CCalendarTime &pTime)
{
    return encode(CDateTime(pTime).GetISODateTime());
}

template<>
CCalendarTime semantic_cast<CCalendarTime,std::wstring>(const std::wstring &pStr)
{
    return semantic_cast<CCalendarTime>(encode(pStr));
}

template<>
std::wstring semantic_cast<std::wstring,CCalendarTime>(const CCalendarTime &pTime)
{
    return CDateTime(pTime).GetISODateTime();
}

///\todo
/*
//////////////////////////////// utility classes /////////////////////////////////

//////////
// uuid
//

template<>
uuid semantic_cast<uuid,std::string>(const std::string &pStr)
{
    std::istringstream lStr(pStr);
    uuid lUUID;
    lStr >> lUUID;
    return lUUID;
}

template<>
std::string semantic_cast<std::string,uuid>(const uuid &pUUID)
{
    std::ostringstream lStr;
    lStr << pUUID;
    return lStr.str();
}

template<>
uuid semantic_cast<uuid,std::wstring>(const std::wstring &pStr)
{
    std::wistringstream lStr(pStr);
    uuid lUUID;
    lStr >> lUUID;
    return lUUID;
}

template<>
std::wstring semantic_cast<std::wstring,uuid>(const uuid &pUUID)
{
    std::wostringstream lStr;
    lStr << pUUID;
    return lStr.str();
}
*/
//////////////////////// strings and vectors of strings //////////////////////////

/////////////////////
// between strings
//

template<>
std::string semantic_cast<std::string,std::string>(const std::string &pStr)
{
    return pStr;
}

template<>
std::string semantic_cast<std::string,std::wstring>(const std::wstring &pStr)
{
    return encode(pStr);
}

template<>
std::wstring semantic_cast<std::wstring,std::wstring>(const std::wstring &pStr)
{
    return pStr;
}

template<>
std::wstring semantic_cast<std::wstring,std::string>(const std::string &pStr)
{
    return decode(pStr);
}

//////////////////////
// vector<std::wstring>
//

template<>
std::vector<std::wstring> semantic_cast<std::vector<std::wstring> ,std::string>(const std::string& /*pStr*/)
{
    throw XNotImplemented("semantic_cast<stringlist,string>");
}

template<>
std::string semantic_cast<std::string,std::vector<std::wstring>>(const std::vector<std::wstring>& /*pV*/)
{
    throw XNotImplemented("semantic_cast<stringlist,string>");
}

template<>
std::vector<std::wstring> semantic_cast<std::vector<std::wstring> ,std::wstring>(const std::wstring& /*pStr*/)
{
    throw XNotImplemented("semantic_cast<stringlist,string>");
}

template<>
std::wstring semantic_cast<std::wstring,std::vector<std::wstring>>(const std::vector<std::wstring>& /*pV*/)
{
    throw XNotImplemented("semantic_cast<stringlist,string>");
}

//////////////////////
// vector<std::string>
//

template<>
std::vector<std::string> semantic_cast<std::vector<std::string> ,std::string>(const std::string& /*pStr*/)
{
    throw XNotImplemented("semantic_cast<stringlist,string>");
}

template<>
std::string semantic_cast<std::string,std::vector<std::string>>(const std::vector<std::string>& /*pV*/)
{
    throw XNotImplemented("semantic_cast<stringlist,string>");
}

template<>
std::vector<std::string> semantic_cast<std::vector<std::string> ,std::wstring>(const std::wstring& /*pStr*/)
{
    throw XNotImplemented("semantic_cast<stringlist,string>");
}

template<>
std::wstring semantic_cast<std::wstring,std::vector<std::string>>(const std::vector<std::string>& /*pV*/)
{
    throw XNotImplemented("semantic_cast<stringlist,string>");
}

/*namespace buola*/ }
