//////////////////////////////////////////////////////////////////////////////////
//                                                                              //
//  This file is part of the buola project (https://code.google.com/p/buola/).  //
//                                                                              //
//  Copyright(c) 2007-2012 Xavi Gratal                                          //
//  gratal AT gmail DOT com                                                     //
//                                                                              //
//  Buola is free software: you can redistribute it and/or modify               //
//  it under the terms of the GNU General Public License as published by        //
//  the Free Software Foundation, either version 3 of the License, or           //
//  (at your option) any later version.                                         //
//                                                                              //
//  Buola is distributed in the hope that it will be useful,                    //
//  but WITHOUT ANY WARRANTY; without even the implied warranty of              //
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the               //
//  GNU General Public License for more details.                                //
//                                                                              //
//  You should have received a copy of the GNU General Public License           //
//  along with buola.  If not, see <http://www.gnu.org/licenses/>.              //
//                                                                              //
//////////////////////////////////////////////////////////////////////////////////

#include <buola/buola.h>

#include <iconv.h>
#include <cerrno>
#include <cwchar>
#include <map>
#include <libintl.h>

namespace buola {

using std::lock;
using std::unique_lock;

std::wstring decode(const char *pSrc,size_t pLen,const char *pCod)
{
    static std::map<const char*,iconv_t> sDecodeMap;
    static mutex sDecodeMutex;

    if(!pSrc) return std::wstring();

    wchar_t lBuffer[pLen+1];

    sDecodeMutex.lock();

    iconv_t lIConv;
    auto lConvIt=sDecodeMap.find(pCod);
    if(lConvIt==sDecodeMap.end())
    {
        lIConv=iconv_open("WCHAR_T",pCod);
        sDecodeMap[pCod]=lIConv;
    }
    else
    {
        lIConv=lConvIt->second;
        iconv(lIConv,nullptr,nullptr,nullptr,nullptr);
    }

    std::size_t lInC=pLen;
    std::size_t lOutC=pLen*sizeof(wchar_t);

    char *lInP=const_cast<char*>(pSrc);
    char *lOutP=(char*)lBuffer;

    if(iconv(lIConv,&lInP,&lInC,&lOutP,&lOutC)==(std::size_t)-1)
    {
        sDecodeMutex.unlock();
        if(!strcmp(pCod,COD_UTF8))
        {
            return decode(pSrc,COD_ISO8859_1);
        }
        else if(!strcmp(pCod,COD_ISO8859_1))
        {
            msg_warn() << "invalid encoding in decode\n";
            return decode(pSrc,COD_ASCII);
        }
        for(int i=0;i<=pLen;i++)
        {
            lBuffer[i]=(wchar_t)pSrc[i];
        }

        return lBuffer;
    }

    lBuffer[pLen-lOutC/4]=0;

    sDecodeMutex.unlock();
    return lBuffer;
}

std::wstring decode(const char *pSrc,const char *pCod)
{
    if(!pSrc) return std::wstring();

    std::size_t lLen;

    for(lLen=0;pSrc[lLen];lLen++) {}

    return decode(pSrc,lLen,pCod);
}

std::string encode(const wchar_t *pSrc,size_t pLen,const char *pCod)
{
    static std::map<const char*,iconv_t> sEncodeMap;
    static mutex sEncodeMutex;

    if(!pSrc) return std::string();
    
    int lBufferLen=pLen*6+1;
    char lBuffer[lBufferLen];

    unique_lock<mutex> lLock(sEncodeMutex);

    iconv_t lIConv;
    auto lConvIt=sEncodeMap.find(pCod);
    if(lConvIt==sEncodeMap.end())
    {
        lIConv=iconv_open(pCod,"WCHAR_T");
        sEncodeMap[pCod]=lIConv;
    }
    else
    {
        lIConv=lConvIt->second;
        iconv(lIConv,nullptr,nullptr,nullptr,nullptr);
    }

    std::size_t lInC=pLen*sizeof(wchar_t);
    std::size_t lOutC=lBufferLen;

    char *lInP=const_cast<char*>((const char*)pSrc);
    char *lOutP=(char*)lBuffer;

    if(iconv(lIConv,&lInP,&lInC,&lOutP,&lOutC)==(std::size_t)-1)
    {
        msg_info() << "can't encode!!!!!!!\n";
        back_trace();
        throw XInternal(I18N("can't do string encoding"));
    }

    lBuffer[lBufferLen-lOutC]=0;

    return lBuffer;
}

std::string encode(const wchar_t *pSrc,const char *pCod)
{
    if(!pSrc) return std::string();

    std::size_t lLen;

    for(lLen=0;pSrc[lLen];lLen++) {}

    return encode(pSrc,lLen,pCod);
}

std::string transcode(const char *pSrc,size_t pLen,const char *pFrom,const char *pTo)
{
    if(!pSrc) return std::string();

    int lBufferLen=pLen*6+1;
    char lBuffer[lBufferLen];

    iconv_t lIConv=iconv_open(pTo,pFrom);

    std::size_t lInC=pLen;
    std::size_t lOutC=lBufferLen;

    char *lInP=const_cast<char*>(pSrc);
    char *lOutP=(char*)lBuffer;

    if(iconv(lIConv,&lInP,&lInC,&lOutP,&lOutC)==(std::size_t)-1)
    {
        if(!strcmp(pFrom,COD_UTF8))
        {
            return transcode(pSrc,COD_ISO8859_1,pTo);
        }
        else if(!strcmp(pFrom,COD_ISO8859_1))
        {
            msg_warn() << "invalid encoding in transcode\n";
            return transcode(pSrc,COD_ASCII,pTo);
        }
        for(int i=0;i<=pLen;i++)
        {
            lBuffer[i]=(wchar_t)pSrc[i];
        }

        return lBuffer;
    }

    iconv_close(lIConv);

    lBuffer[lBufferLen-lOutC]=0;

    return lBuffer;
}

std::string transcode(const char *pSrc,const char *pFrom,const char *pTo)
{
    if(!pSrc) return std::string();

    std::size_t lLen;

    for(lLen=0;pSrc[lLen];lLen++) {}

    return transcode(pSrc,lLen,pFrom,pTo);
}

std::string I18N(const char *pStr)
{
    return gettext(pStr);
}

std::wstring WI18N(const char *pStr)
{
    return decode(I18N(pStr));
}

static const char sHexDigits[16]={ '0','1','2','3','4','5','6','7',
                                    '8','9','a','b','c','d','e','f' };

std::string hex_string(const uint8_t *pData,size_t pLen)
{
    std::string lReturn;
    lReturn.resize(pLen*2);

    for(int i=0;i<pLen;i++)
    {
        lReturn[i*2]=sHexDigits[pData[i]>>4];
        lReturn[i*2+1]=sHexDigits[pData[i]&0x0f];
    }

    return lReturn;
    
}

std::wstring hex_wstring(const uint8_t *pData,size_t pLen)
{
    std::wstring lReturn;
    lReturn.resize(pLen*2);

    for(int i=0;i<pLen;i++)
    {
        lReturn[i*2]=sHexDigits[pData[i]>>4];
        lReturn[i*2+1]=sHexDigits[pData[i]&0x0f];
    }

    return lReturn;
}

/*namespace buola*/ }
